#!/usr/bin/env bash
indent -nbad -bap -nbc -bbo -hnl -br -brs -c33 -cd33 -ncdb -ce -ci4 \
-cli0 -d0 -di1 -nfc1 -i4 -ip0 -l80 -lp -npcs -nprs -npsl -ppi4 -sai \
-saf -saw -ncs -nsc -sob -nfca -cp33 -ss -ts8 -il1 -nut -brf $@
