#pragma once
#ifndef LICENCE_H
#    define LICENCE_H

static const char *scrappie_licence_text =
    "This software is subject to the terms of the Mozilla Public\n"
    "License, v. 2.0. If a copy of the MPL was not distributed with this\n"
    "file, You can obtain one at http://mozilla.org/MPL/2.0/.\n"
    "\n"
    "(c) 2017 Oxford Nanopore Technologies Ltd.\n"
    "\n"
    "\n"
    "\nThe vectorised math functions 'src/sse_mathfun.h' are from\n"
    "http://gruntthepeon.free.fr/ssemath/ and the original version of this file is\n"
    "under the 'zlib' licence.  See the top of 'src/sse_mathfun.h' for details.\n";

#endif                          /* LICENCE_H */
