#include <stdio.h>

#include "decode.h"
#include "event_detection.h"
#include "fast5_interface.h"
#include "homopolymer.h"
#include "kseq.h"
#include "layers.h"
#include "networks.h"
#include "nnfeatures.h"
#include "scrappie_common.h"
#include "scrappie_matrix.h"
#include "scrappie_seq_helpers.h"
#include "scrappie_stdlib.h"
#include "scrappie_structures.h"
#include "scrappie_subcommands.h"
#include "sse_mathfun.h"
#include "util.h"

#include "../python/pyscrap.h"


int main(void){
    return EXIT_SUCCESS;
}
